from .base import *
from .keyword import *
from .keyword_preset import *
from .mention_spam import *
from .utils import *

__all__ = (
    *base.__all__,
    *keyword.__all__,
    *keyword_preset.__all__,
    *mention_spam.__all__,
    *utils.__all__,
)



__all__ = ('AutoModerationAction',)

from scarletio import RichAttributeErrorBaseType

from ..preconverters import preconvert_preinstanced_type

from .preinstanced import AutoModerationActionType


def _validate_action_type_with_metadata_options(action_type, channel, duration):
    """
    Validates the given `action_type` with the `channel`, `duration` options. If any option is given, the
    `action_type` will default towards it. On any mismatch exception is raised.
    
    Parameters
    ----------
    action_type : `Ellipsis`, `int`, ``AutoModerationActionType``
        Auto moderation action type.
    
    channel : `Ellipsis`, `None`, ``Channel``, `int`
        The channel where the alert messages should be sent.

    duration : `Ellipsis`, `None`, `int`, `float`
        The timeout's duration applied on trigger.
    
    Returns
    -------
    action_metadata : `None`, ``AutoModerationActionMetadataBase``
        Action type specific metadata if applicable.
    action_type : ``AutoModerationActionType``
        The final processed action type.
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
        - If there are multiple mutually exclusive options.
        - If action type not detectable.
        - If multiple action types detected.
    ValueError
        - If a parameter's value is incorrect.
    """
    if (action_type is not ...):
        action_type = preconvert_preinstanced_type(action_type, 'action_type', AutoModerationActionType)
    
    if (channel is not ...) + (duration is not ...) > 1:
        raise TypeError(
            f'`channel` and `duration` parameters are mutually exclusive, got '
            f'channel = {channel!r}; duration = {duration!r}.'
        )
    
    if (channel is not ...):
        probable_action_type = AutoModerationActionType.send_alert_message
        metadata_parameter = channel
    
    elif (duration is not ...):
        probable_action_type = AutoModerationActionType.timeout
        metadata_parameter = duration
    
    else:
        probable_action_type = None
        metadata_parameter = None
    
    if (action_type is ...):
        if (probable_action_type is None):
            raise TypeError(
                f'`action_type` is not given or is given as `None`, and cannot be detected from `channel` '
                'or from `duration` parameters.'
            )
        
        action_type = probable_action_type
    
    else:
        if (probable_action_type is not None):
            if (action_type is not probable_action_type):
                if (channel is not ...):
                    received_parameter_name = 'channel'
                    received_parameter_value = channel
                    
                else:
                    received_parameter_name = 'duration'
                    received_parameter_value = duration
                
                raise TypeError(
                    f'Both `action_type` and `{received_parameter_name}` parameters refer to a different '
                    f'action type, got action_type = {action_type!r}, '
                    f'{received_parameter_name} = {received_parameter_value!r}.'
                )
    
    action_metadata_type = action_type.metadata_type
    if (action_metadata_type is None):
        action_metadata = None
    
    else:
        action_metadata = action_metadata_type(metadata_parameter)
    
    return action_metadata, action_type


class AutoModerationAction(RichAttributeErrorBaseType):
    """
    An action that is executed when a rule is triggered.
    
    Attributes
    ----------
    metadata : `None`, ``AutoModerationActionMetadataBase``
        Characterises the action
    type : ``AutoModerationActionType``
        The action's type.
    """
    __slots__ = ('metadata', 'type')
    
    def __new__(cls, action_type=..., *, channel = ..., duration=...):
        """
        Creates a new auto moderation action from the given parameters.
        
        Parameters
        ----------
        action_type : ``AutoModerationActionType``, `int`, Optional
            The moderation action's type.
        
        channel : `None`, ``Channel``, `int`, Optional (Keyword only)
            The channel where the alert messages should be sent.
            
            > Mutually exclusive with the `duration` parameter.
        
        duration : `None`, `int`, `float`, Optional (Keyword only)
            The timeout's duration applied on trigger.
            
            > Mutually exclusive with the `channel` parameter.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Parameter mismatch.
        """
        # type & channel & duration
        metadata, type_ = _validate_action_type_with_metadata_options(action_type, channel, duration)
        
        self = object.__new__(cls)
        self.type = type_
        self.metadata = metadata
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``AutoModerationAction`` from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received auto moderation action data.
        
        Returns
        -------
        self : ``AutoModerationAction``
            The created auto moderation rule.
        """
        # type_
        type_ = AutoModerationActionType.get(data['type'])
        
        metadata_type = type_.metadata_type
        if (metadata_type is None):
            metadata = None
        
        else:
            metadata = metadata_type.from_data(data['metadata'])
        
        self = object.__new__(cls)
        self.type = type_
        self.metadata = metadata
        return self
    
    
    def to_data(self):
        """
        Converts the auto moderation action to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # type
        data['type'] = self.type.value
        
        # metadata
        metadata = self.metadata
        if metadata is None:
            metadata_data = {}
        else:
            metadata_data = metadata.to_data()
        
        data['metadata'] = metadata_data
        
        return data
    
    
    def __eq__(self, other):
        """Returns whether the two auto moderation actions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.type is not other.type:
            return False
        
        if self.metadata != other.metadata:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the auto moderation action's hash value."""
        hash_value = 0
        
        hash_value ^= self.type.value
        
        metadata = self.metadata
        if (metadata is not None):
            hash_value ^= hash(metadata)
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the auto moderation action's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # type
        type_ = self.type
        
        repr_parts.append(' type=')
        repr_parts.append(repr(type_.name))
        repr_parts.append('~')
        repr_parts.append(repr(type_.value))
        
        # metadata
        metadata = self.metadata
        if (metadata is not None):
            repr_parts.append(', metadata=')
            repr_parts.append(repr(metadata))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the auto moderation action.
        
        Returns
        -------
        new : ``AutoModerationAction``
        """
        new = object.__new__(type(self))
        
        # metadata
        metadata = self.metadata
        if (metadata is not None):
            metadata = metadata.copy()
        new.metadata = metadata
        
        # type
        new.type = self.type
        
        return new

    
    def copy_with(self, **keyword_parameters):
        """
        Copies the auto moderation action with the given attributes replaced.
        
        Parameters
        ----------
        action_type : ``AutoModerationActionType``, `int`, Optional (Keyword only)
            The moderation action's type.
        
        channel : `None`, ``Channel``, `int`, Optional (Keyword only)
            The channel where the alert messages should be sent.
            
            > Mutually exclusive with the `duration` parameter.
        
        duration : `None`, `int`, `float`, Optional (Keyword only)
            The timeout's duration applied on trigger.
            
            > Mutually exclusive with the `channel` parameter.
        
        Returns
        -------
        self : ``AutoModerationAction``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Parameter mismatch.
        """
        # trigger_metadata & trigger_type
        action_type = keyword_parameters.pop('action_type', ...)
        channel = keyword_parameters.pop('channel', ...)
        duration = keyword_parameters.pop('duration', ...)
        
        if (action_type is ...) and (channel is ...) and (duration is ...):
            metadata = self.metadata
            if (metadata is not None):
                metadata = metadata.copy()
            
            type_ = self.type
            
        else:
            metadata, type_ = _validate_action_type_with_metadata_options(
                action_type, channel, duration
            )
            
        self = object.__new__(type(self))
        self.type = type_
        self.metadata = metadata
        return self



__all__ = ()

from ...permission import Permission
from ...utils import PERMISSION_DENY_KEY


def parse_deny(data):
    """
    Parses the `deny` field of ``PermissionOverwrite`` out of the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Permission overwrite data.
    
    Returns
    -------
    deny : ``Permission``
    """
    return Permission(data[PERMISSION_DENY_KEY])


def validate_deny(deny):
    """
    Validates `deny` field of ``PermissionOverwrite``.
    
    Parameters
    ----------
    deny : `None`, ``Permission``, `int`
        The permission overwrite's denied permission's value.
    
    Returns
    -------
    deny : ``Permission``
    
    Raises
    ------
    TypeError
        - If `deny` is not `None`, `int`, ``Permission``.
    """
    if deny is None:
        deny = Permission()
    
    elif isinstance(deny, Permission):
        pass
    
    elif isinstance(deny, int):
        deny = Permission(deny)
    
    else:
        raise TypeError(
            f'`deny` can be `None`, `{Permission.__name__}`, `int`, got {deny.__class__.__name__}; {deny!r}.'
        )
    
    return deny


def put_deny_into(deny, data, defaults):
    """
    Puts the `deny` field of ``PermissionOverwrite`` to the given data.
    
    Parameters
    ----------
    deny : ``Permission``
        The permission overwrite's denied permission's value.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    data[PERMISSION_DENY_KEY] = format(deny, 'd')
    
    return data



__all__ = ('ComponentMetadataRow', )

from scarletio import copy_docs

from ..shared_fields import parse_components, put_components_into, validate_components

from .base import ComponentMetadataBase


class ComponentMetadataRow(ComponentMetadataBase):
    """
    Row component metadata.
    
    Attributes
    ----------
    components : `None`, `tuple` of ``Component``
        The contained components.
    """
    __slots__ = ('components',)
    
    @copy_docs(ComponentMetadataBase.__new__)
    def __new__(cls, keyword_parameters):
        # components
        try:
            components = keyword_parameters.pop('components')
        except KeyError:
            components = None
        else:
            components = validate_components(components)
        
        self = object.__new__(cls)
        self.components = components
        return self
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # sub-component fields : components
        
        # components
        repr_parts.append(' components=')
        components = self.components
        if (components is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(components)
            
            while True:
                component = components[index]
                index += 1
                
                repr_parts.append(repr(component))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components) << 12
            for component in components:
                hash_value ^= hash(component)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # components
        if self.components != other.components:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        self.components = parse_components(data)
        
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data =  {}
        
        put_components_into(self.components, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # components
        components = self.components
        if (components is not None):
            components = tuple(component.copy() for component in self.components)
        
        new.components = components
        
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with)
    def copy_with(self, keyword_parameters):
        # components
        try:
            components = keyword_parameters.pop('components')
        except KeyError:
            components = self.components
            if (components is not None):
                components = tuple(component.copy() for component in self.components)
        else:
            components = validate_components(components)
        
        # Construct
        
        self = object.__new__(type(self))
        self.components = components
        return self



__all__ = ('Connection', )

from scarletio import include

from ...bases import DiscordEntity
from ...preconverters import preconvert_snowflake

from .fields import (
    parse_friend_sync, parse_integrations, parse_name, parse_revoked, parse_show_activity, parse_two_way_link,
    parse_type, parse_verified, parse_visibility, put_friend_sync_into, put_integrations_into, put_name_into,
    put_revoked_into, put_show_activity_into, put_two_way_link_into, put_type_into, put_verified_into,
    put_visibility_into, validate_friend_sync, validate_integrations, validate_name, validate_revoked,
    validate_show_activity, validate_two_way_link, validate_type, validate_verified, validate_visibility
)
from .preinstanced import ConnectionType, ConnectionVisibility


Integration = include('Integration')


class Connection(DiscordEntity):
    """
    A connection object that a user is attached to.
    
    Attributes
    ----------
    id : `int`
        The unique identifier value of the connection.
    
    friend_sync : `bool`
        Whether the user has friend sync enabled for the connection.
    
    integrations : `None` or `tuple` of ``Integration``
        A guild's integrations which are attached to the connection.
    
    name : `str`
        The username of the connected account.
    
    revoked : `bool`
        Whether the connection is revoked.
    
    show_activity : `bool`
        Whether activity related to this connection will be shown in presence updates.
    
    two_way_link : `bool`
         Whether this connection has a corresponding third party OAuth2 token.
    
    type : ``ConnectionType``
        The service of the connection.
    
    verified : `bool`
        Whether the connection is verified.
    
    visibility : ``ConnectionVisibility``
        For who is the connection visible for.
    """
    __slots__ = (
        'friend_sync', 'integrations', 'name', 'revoked', 'show_activity', 'two_way_link', 'type', 'verified',
        'visibility'
    )
    
    
    def __new__(cls, **keyword_parameters):
        """
        Creates a new connection from the given parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword Parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        friend_sync : `bool`, Optional (Keyword only)
            Whether the user has friend sync enabled for the connection.
        
        integrations : `None` or `iterable` of ``Integration``, Optional (Keyword only)
            A guild's integrations which are attached to the connection.
        
        name : `str`, Optional (Keyword only)
            The username of the connected account.
        
        revoked : `bool`, Optional (Keyword only)
            Whether the connection is revoked.
        
        show_activity : `bool`, Optional (Keyword only)
            Whether activity related to this connection will be shown in presence updates.
        
        two_way_link : `bool`, Optional (Keyword only)
             Whether this connection has a corresponding third party OAuth2 token.
        
        type : ``ConnectionType``, `str`, Optional (Keyword only)
            The service of the connection.
        
        verified : `bool`, Optional (Keyword only)
            Whether the connection is verified.
        
        visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection visible for.
        
        Raises
        ------
        TypeError
            - If a field's type is unacceptable.
            - Extra or unused parameters
        ValueError
            - If a a field's value is unacceptable.
        """
        self = cls._create_empty(0)
        self._set_attributes_from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        return self
    
    
    @classmethod
    def precreate(cls, connection_id, **keyword_parameters):
        """
        Creates a new connection with the given predefined fields.
        
        > Since connections are not cached globally, this method is only be used for testing.
        
        Parameters
        ----------
        connection_id : `int`
            The connection's identifier.
        **keyword_parameters : Keyword Parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        friend_sync : `bool`, Optional (Keyword only)
            Whether the user has friend sync enabled for the connection.
        
        integrations : `None` or `iterable` of ``Integration``, Optional (Keyword only)
            A guild's integrations which are attached to the connection.
        
        name : `str`, Optional (Keyword only)
            The username of the connected account.
        
        revoked : `bool`, Optional (Keyword only)
            Whether the connection is revoked.
        
        show_activity : `bool`, Optional (Keyword only)
            Whether activity related to this connection will be shown in presence updates.
        
        two_way_link : `bool`, Optional (Keyword only)
             Whether this connection has a corresponding third party OAuth2 token.
        
        type : ``ConnectionType``, `str`, Optional (Keyword only)
            The service of the connection.
        
        verified : `bool`, Optional (Keyword only)
            Whether the connection is verified.
        
        visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection visible for.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If `connection_id` is not `int`.
            - If a field's type is unacceptable.
            - Extra or unused parameters
        ValueError
            - If a a field's value is unacceptable.
        """
        connection_id = preconvert_snowflake(connection_id, 'connection_id')
        self = cls._create_empty(connection_id)
        self._set_attributes_from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        return self
    
    
    @classmethod
    def _create_empty(cls, connection_id):
        """
        Creates a new connection with it's default attributes set.
        
        Parameters
        ----------
        connection_id : `int`
            The connection's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.id = connection_id
        self.friend_sync = False
        self.integrations = None
        self.name = ''
        self.revoked = False
        self.show_activity = False
        self.two_way_link = False
        self.type = ConnectionType.unknown
        self.verified = False
        self.visibility = ConnectionVisibility.user_only
        
        return self
    
    
    def __repr__(self):
        """Returns the connection's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        connection_id = self.id
        if connection_id:
            repr_parts.append(' id=')
            repr_parts.append(repr(connection_id))
        
        else:
            repr_parts.append(' (partial)')
        
        connection_type = self.type
        if (connection_type is not ConnectionType.unknown):
            repr_parts.append(', type=')
            repr_parts.append(connection_type.name)
            repr_parts.append('~')
            repr_parts.append(repr(connection_type.value))
        
        name = self.name
        if name:
            repr_parts.append(', name=')
            repr_parts.append(repr(name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two connections are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two connections are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id == other_id:
                return True
            
            return False
        
        # friend_sync
        if self.friend_sync != other.friend_sync:
            return False
        
        # integrations
        if self.integrations != other.integrations:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # revoked
        if self.revoked != other.revoked:
            return False
        
        # show_activity
        if self.show_activity != other.show_activity:
            return False
        
        # two_way_link
        if self.two_way_link != other.two_way_link:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # verified
        if self.verified != other.verified:
            return False
        
        # visibility
        if self.visibility is not other.visibility:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a connection object from received connection data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`)
            Received connection data.
        """
        self = object.__new__(cls)
        
        self.id = int(data['id'])
        self.friend_sync = parse_friend_sync(data)
        self.integrations = parse_integrations(data)
        self.name = parse_name(data)
        self.revoked = parse_revoked(data)
        self.show_activity = parse_show_activity(data)
        self.two_way_link = parse_two_way_link(data)
        self.type = parse_type(data)
        self.verified = parse_verified(data)
        self.visibility = parse_visibility(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the connection to json serializable representation dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether we want to include identifiers as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = {}
        
        if include_internals:
            data['id'] = str(self.id)
        
        put_friend_sync_into(self.friend_sync, data, defaults)
        put_integrations_into(self.integrations, data, defaults, include_internals = include_internals)
        put_name_into(self.name, data, defaults)
        put_revoked_into(self.revoked, data, defaults)
        put_show_activity_into(self.show_activity, data, defaults)
        put_two_way_link_into(self.two_way_link, data, defaults)
        put_type_into(self.type, data, defaults)
        put_verified_into(self.verified, data, defaults)
        put_visibility_into(self.visibility, data, defaults)
        
        return data
    
    
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        """
        Sets the connection's attributes from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            The keyword parameters to set the connection's attributes from.
        
        Raises
        ------
        TypeError
            A field of invalid type.
        ValueError
            A field of invalid value.
        """
        # id
        # Ignore, internal
        
        # friend_sync
        try:
            friend_sync = keyword_parameters.pop('friend_sync')
        except KeyError:
            pass
        else:
            self.friend_sync = validate_friend_sync(friend_sync)
        
        # integrations
        try:
            integrations = keyword_parameters.pop('integrations')
        except KeyError:
            pass
        else:
            self.integrations = validate_integrations(integrations)
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            pass
        else:
            self.name = validate_name(name)
        
        # revoked
        try:
            revoked = keyword_parameters.pop('revoked')
        except KeyError:
            pass
        else:
            self.revoked = validate_revoked(revoked)
        
        # show_activity
        try:
            show_activity = keyword_parameters.pop('show_activity')
        except KeyError:
            pass
        else:
            self.show_activity = validate_show_activity(show_activity)
        
        # two_way_link
        try:
            two_way_link = keyword_parameters.pop('two_way_link')
        except KeyError:
            pass
        else:
            self.two_way_link = validate_two_way_link(two_way_link)
        
        # type
        try:
            type_ = keyword_parameters.pop('type_')
        except KeyError:
            pass
        else:
            self.type = validate_type(type_)
        
        # verified
        try:
            verified = keyword_parameters.pop('verified')
        except KeyError:
            pass
        else:
            self.verified = validate_verified(verified)
        
        # visibility
        try:
            visibility = keyword_parameters.pop('visibility')
        except KeyError:
            pass
        else:
            self.visibility = validate_visibility(visibility)
    
    
    def iter_integrations(self):
        """
        Iterates over the integrations of the connection.
        
        This method is an iterable generator.
        
        Yields
        ------
        integration : ``Integration``
        """
        integrations = self.integrations
        if (integrations is not None):
            yield from integrations



from .base import *
from .button import *
from .channel_select import *
from .constants import *
from .fields import *
from .mentionable_select import *
from .preinstanced import *
from .role_select import *
from .row import *
from .select_base import *
from .string_select import *
from .text_input import *
from .user_select import *


__all__ = (
    *base.__all__,
    *button.__all__,
    *channel_select.__all__,
    *constants.__all__,
    *fields.__all__,
    *mentionable_select.__all__,
    *preinstanced.__all__,
    *role_select.__all__,
    *row.__all__,
    *select_base.__all__,
    *string_select.__all__,
    *text_input.__all__,
    *user_select.__all__,
)



__all__ = ('ChannelMetadataGuildForum',)


from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_THREAD_AND_VOICE_DENY

from .constants import AUTO_ARCHIVE_DEFAULT, SLOWMODE_DEFAULT
from .fields import (
    parse_available_tags, parse_default_sort_order, parse_default_thread_auto_archive_after,
    parse_default_thread_reaction, parse_default_thread_slowmode, parse_flags, parse_topic, put_available_tags_into,
    put_default_sort_order_into, put_default_thread_auto_archive_after_into, put_default_thread_reaction_into,
    put_default_thread_slowmode_into, put_flags_into, put_topic_into, validate_available_tags,
    validate_default_sort_order, validate_default_thread_auto_archive_after, validate_default_thread_reaction,
    validate_default_thread_slowmode, validate_flags, validate_topic
)
from .flags import ChannelFlag
from .preinstanced import SortOrder

from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildForum(ChannelMetadataGuildMainBase):
    """
    Guild forum channel metadata.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    available_tags : `None`, `tuple` of ``ForumTag``
        The available tags to assign to the child-thread channels.
    default_sort_order : ``SortOrder``
        How the posts ordered in a forum channel by default.
    default_thread_auto_archive_after : `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Defaults
        to `3600`. Can be one of: `3600`, `86400`, `259200`, `604800`.
    default_thread_reaction : `None`, ``Emoji``
        The emoji to show in the add reaction button on a thread of the forum channel.
    default_thread_slowmode : `int`
        The default slowmode applied to the channel's threads.
    flags : ``ChannelFlag``
        The channel's flags.
    topic : `None`, `str`
        The channel's topic.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = (
        'available_tags', 'default_sort_order', 'default_thread_auto_archive_after', 'default_thread_reaction',
        'default_thread_slowmode', 'flags', 'topic',
    )
    
    @copy_docs(ChannelMetadataGuildMainBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildMainBase.__hash__(self)
        
        # available_tags
        available_tags = self.available_tags
        if (available_tags is not None):
            hash_value ^= len(available_tags) << 4
            
            for available_tag in available_tags:
                hash_value ^= hash(available_tag)
        
        # default_sort_order
        hash_value ^= self.default_sort_order.value << 20
        
        # default_thread_auto_archive_after
        hash_value ^= self.default_thread_auto_archive_after << 16
        
        # default_thread_reaction
        default_thread_reaction = self.default_thread_reaction
        if (default_thread_reaction is not None):
            hash_value ^= hash(default_thread_reaction)
        
        # default_thread_slowmode
        hash_value ^= self.default_thread_slowmode << 8
        
        # flags
        hash_value ^= self.flags
        
        # topic
        topic = self.topic
        if (topic is not None):
            hash_value ^= hash(topic)
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildMainBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildMainBase._is_equal_same_type(self, other):
            return False
        
        # available_tags
        if self.available_tags != other.available_tags:
            return False
        
        # default_sort_order
        if self.default_sort_order is not other.default_sort_order:
            return False
        
        # default_thread_auto_archive_after
        if self.default_thread_auto_archive_after != other.default_thread_auto_archive_after:
            return False
        
        # default_thread_reaction
        if self.default_thread_reaction != other.default_thread_reaction:
            return False
        
        # default_thread_slowmode
        if self.default_thread_slowmode != other.default_thread_slowmode:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # topic
        if self.topic != other.topic:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_display_name)
    def _get_display_name(self):
        return self.name.upper()
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildForum, cls)._create_empty()
        
        self.available_tags = None
        self.default_sort_order = SortOrder.latest_activity
        self.default_thread_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        self.default_thread_reaction = None
        self.default_thread_slowmode = SLOWMODE_DEFAULT
        self.flags = ChannelFlag()
        self.topic = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        # available_tags
        self.available_tags = parse_available_tags(data)
        
        # default_sort_order
        self.default_sort_order = parse_default_sort_order(data)
        
        # default_thread_auto_archive_after
        self.default_thread_auto_archive_after = parse_default_thread_auto_archive_after(data)
        
        # default_thread_reaction
        self.default_thread_reaction = parse_default_thread_reaction(data)
        
        # default_thread_slowmode
        self.default_thread_slowmode = parse_default_thread_slowmode(data)
        
        # flags
        self.flags = parse_flags(data)
        
        # topic
        self.topic = parse_topic(data)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        # available_tags
        available_tags = parse_available_tags(data)
        if (self.available_tags != available_tags):
            old_attributes['available_tags'] = self.available_tags
            self.available_tags = available_tags
        
        # default_sort_order
        default_sort_order = parse_default_sort_order(data)
        if self.default_sort_order is not default_sort_order:
            old_attributes['default_sort_order'] = self.default_sort_order
            self.default_sort_order = default_sort_order
        
        # default_thread_auto_archive_after
        default_thread_auto_archive_after = parse_default_thread_auto_archive_after(data)
        if self.default_thread_auto_archive_after != default_thread_auto_archive_after:
            old_attributes['default_thread_auto_archive_after'] = self.default_thread_auto_archive_after
            self.default_thread_auto_archive_after = default_thread_auto_archive_after
        
        # default_thread_reaction
        default_thread_reaction = parse_default_thread_reaction(data)
        if self.default_thread_reaction != default_thread_reaction:
            old_attributes['default_thread_reaction'] = self.default_thread_reaction
            self.default_thread_reaction = default_thread_reaction
        
        # default_thread_slowmode
        default_thread_slowmode = parse_default_thread_slowmode(data)
        if self.default_thread_slowmode != default_thread_slowmode:
            old_attributes['default_thread_slowmode'] = self.default_thread_slowmode
            self.default_thread_slowmode = default_thread_slowmode
        
        # flags
        flags = parse_flags(data)
        if (self.flags != flags):
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        # topic
        topic = parse_topic(data)
        if self.topic != topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        ChannelMetadataGuildMainBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
        
        # available_tags
        try:
            available_tags = keyword_parameters.pop('available_tags')
        except KeyError:
            pass
        else:
            self.available_tags = validate_available_tags(available_tags)
        
        # default_sort_order
        try:
            default_sort_order = keyword_parameters.pop('default_sort_order')
        except KeyError:
            pass
        else:
            self.default_sort_order = validate_default_sort_order(default_sort_order)
        
        # default_thread_auto_archive_after
        try:
            default_thread_auto_archive_after = keyword_parameters.pop('default_thread_auto_archive_after')
        except KeyError:
            pass
        else:
            self.default_thread_auto_archive_after = validate_default_thread_auto_archive_after(
                default_thread_auto_archive_after
            )
        
        # default_thread_reaction
        try:
            default_thread_reaction = keyword_parameters.pop('default_thread_reaction')
        except KeyError:
            pass
        else:
            self.default_thread_reaction = validate_default_thread_reaction(default_thread_reaction)
        
        # default_thread_slowmode
        try:
            default_thread_slowmode = keyword_parameters.pop('default_thread_slowmode')
        except KeyError:
            pass
        else:
            self.default_thread_slowmode = validate_default_thread_slowmode(default_thread_slowmode)
        
        # flags
        try:
            flags = keyword_parameters.pop('flags')
        except KeyError:
            pass
        else:
            self.flags = validate_flags(flags)
        
        # topic
        try:
            topic = keyword_parameters.pop('topic')
        except KeyError:
            pass
        else:
            self.topic = validate_topic(topic)
    
    
    @copy_docs(ChannelMetadataGuildMainBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildMainBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # available_tags
        put_available_tags_into(self.available_tags, data, defaults, include_internals = include_internals)
        
        # default_sort_order
        put_default_sort_order_into(self.default_sort_order, data, defaults)
        
        # default_auto_archive_duration
        put_default_thread_auto_archive_after_into(self.default_thread_auto_archive_after, data, defaults)
        
        # default_thread_reaction
        put_default_thread_reaction_into(self.default_thread_reaction, data, defaults)
        
        # default_thread_slowmode
        put_default_thread_slowmode_into(self.default_thread_slowmode, data, defaults)
        
        # flags
        put_flags_into(self.flags, data, defaults)
        
        # topic
        put_topic_into(self.topic, data, defaults)
        
        return data



__all__ = ('InteractionMetadataBase',)

from scarletio import RichAttributeErrorBaseType, copy_docs

from ...bases import PlaceHolder
from ...component import ComponentType

from ..resolved import Resolved

ENTITY_RESOLVERS = {
    ComponentType.user_select: (lambda resolved, entity_id: resolved.resolve_user(entity_id)),
    ComponentType.role_select: (lambda resolved, entity_id: resolved.resolve_role(entity_id)),
    ComponentType.mentionable_select: (lambda resolved, entity_id: resolved.resolve_mentionable(entity_id)),
    ComponentType.channel_select: (lambda resolved, entity_id: resolved.resolve_channel(entity_id)),
}


class InteractionMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for values assigned to ``InteractionEvent.interaction`` field.
    """
    __slots__ = ()
    
    def __new__(cls, **keyword_parameters):
        """
        Creates a new interaction metadata from the given parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining how should the fields be set.
        
        Other Parameters
        ----------------
        component_type : ``ComponentType``, Optional (Keyword only)
            The used component's type.
        
        components : `None`, `tuple` of ``InteractionComponent``, Optional (Keyword only)
            Submitted component values of a form submit interaction.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Component or form interaction's custom identifier.
        
        id : `int`, Optional (Keyword only)
            The represented application command's identifier number.
        
        name : `str`, Optional (Keyword only)
            The represented application command's name.
        
        options : `None`, `tuple` of ``InteractionOption``, Optional (Keyword only)
            Application command option representations. Like sub-command or parameter.
        
        resolved : `None`, ``Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        target_id : `int`, Optional (Keyword only)
            The interaction's target's identifier. Applicable for context commands.
        
        values : `None`, `tuple` of `str`, Optional (Keyword only)
            Values selected by the user. Applicable for component interactions.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a field's value is incorrect.
        """
        self = cls._create_empty()
        if keyword_parameters:
            self._set_attributes_from_keyword_parameters(keyword_parameters)
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an new interaction with it's attribute set as it's default values.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)    
    
    
    def copy(self):
        """
        Copies the integration option.
        
        Returns
        -------
        new : `instance<cls>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copies the integration with replacing the defined fields.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which fields and how should be set.
        
        Other Parameters
        ----------------
        component_type : ``ComponentType``, Optional (Keyword only)
            The used component's type.
        
        components : `None`, `tuple` of ``InteractionComponent``, Optional (Keyword only)
            Submitted component values of a form submit interaction.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Component or form interaction's custom identifier.
        
        id : `int`, Optional (Keyword only)
            The represented application command's identifier number.
        
        name : `str`, Optional (Keyword only)
            The represented application command's name.
        
        options : `None`, `tuple` of ``InteractionOption``, Optional (Keyword only)
            Application command option representations. Like sub-command or parameter.
        
        resolved : `None`, ``Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        target_id : `int`, Optional (Keyword only)
            The interaction's target's identifier. Applicable for context commands.
        
        values : `None`, `tuple` of `str`, Optional (Keyword only)
            Values selected by the user. Applicable for component interactions.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        self = self.copy()
        if keyword_parameters:
            self._set_attributes_from_keyword_parameters(keyword_parameters)
        return self
    
    
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        """
        Sets the integration's attributes from the given keyword parameters.
        
        Parameters
        keyword_parameters : Keyword parameters
            A dictionary of keyword parameters defining which fields and how should be set.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
    
    
    @classmethod
    def from_data(cls, data, interaction_event):
        """
        Creates a new interaction from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received interaction field data.
        
        interaction_event : ``InteractionEvent``
            The parent interaction event.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(cls, *, defaults = False, interaction_event = None):
        """
        Converts the interaction into a json serailzable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        interaction_event : ``InteractionEvent`` = `None`, Optional (Keyword only)
            The respective guild's identifier to use for handing user guild profiles.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __repr__(self):
        """Returns the interaction's representation."""
        repr_parts = []
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        
        self._put_attribute_representations_into(repr_parts)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_attribute_representations_into(self, repr_parts):
        """
        Helper function to build representation of the interaction metadata.
        
        Parameters
        ----------
        repr_parts : `list` of `str`
            Integration metadata representation parts.
        
        Returns
        -------
        field_added : `bool`
            Whether any field was added.
        """
        return False
    
    
    def __hash__(self):
        """Returns the interaction's hash value."""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two interactions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two interactions are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other interaction. Must be the same type as `self`.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    component_type = PlaceHolder(
        ComponentType.none,
        """
        The used component's type.
        
        Returns
        -------
        component_type : ``ComponentType``
        """
    )
    
    
    components = PlaceHolder(
        None,
        """
        Submitted component values of a form submit interaction.
        
        Returns
        -------
        components : `None`, `tuple` of ``InteractionComponent``
        """
    )
    
    
    custom_id = PlaceHolder(
        None,
        """
        Component or form interaction's custom identifier.
        
        Returns
        -------
        custom_id : `None`, `str`
        """,
    )
    
    
    id = PlaceHolder(
        0,
        """
        The represented application command's identifier number.
        
        Returns
        -------
        application_command_id : `int`
        """,
    )
    
    
    name = PlaceHolder(
        '',
        """
        The represented application command's name.
        
        Returns
        -------
        application_command_name : `str`
        """,
    )
    
    
    options = PlaceHolder(
        None,
        """
        Application command option representations. Like sub-command or parameter.
            
        Returns
        -------
        options : `None`, `tuple` of ``InteractionOption``
        """,
    )
    
    
    resolved = PlaceHolder(
        None,
        """
        Contains the received entities.
        
        Returns
        -------
        resolved : `None`, ``Resolved``
        """,
    )
    
    
    target_id = PlaceHolder(
        0,
        """
        The interaction's target's identifier. Applicable for context commands.
        
        Returns
        -------
        target_id : `int`
        """,
    )
    
    
    values = PlaceHolder(
        None,
        """
        Values selected by the user. Applicable for component interactions.
        
        Returns
        -------
        values : `None`, `tuple` of `str`
        """
    )
    
    # Extra utility | Application command
    
    @property
    def target(self):
        """
        Returns the application command's target. Applicable if the interaction was invoked by a context command.
        
        Returns
        -------
        entity : `None` ``Attachment``, ``Channel``, ``ClientUserBase``, ``Role``, ``Message``
        """
        target_id = self.target_id
        if target_id:
            return self.resolve_entity(target_id)
    
    # Extra utility
    
    # Application command autocomplete
    
    def iter_options(self):
        """
        Iterates over the options of the interaction application command (autocomplete) interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        option : ``InteractionOption``
        """
        options = self.options
        if (options is not None):
            yield from options
    
    
    @property
    def focused_option(self):
        """
        Returns the focused option of the application command autocomplete interaction.
        
        Returns
        -------
        option : `None`, ``InteractionOption``
        """
        for option in self.iter_options():
            focused_option = option.focused_option
            if (focused_option is not None):
                return focused_option
    
    
    def get_non_focused_values(self):
        """
        Gets the non focused values of the interaction.
        
        Returns
        -------
        non_focused_options : `dict` of (`str`, (`None`, `str`)) items
        """
        return dict(self._iter_non_focused_values())
    
    
    def _iter_non_focused_values(self):
        """
        Iterates over the non focused values of the interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
            The option's name.
        
        value : `None`, `str`
            The option's value.
        """
        for option in self.iter_options():
            yield from option._iter_non_focused_values()
    
    
    def get_value_of(self, *option_names):
        """
        Gets the value for the option by the given name.
        
        Parameters
        ----------
        *option_names : `str`
            The option(s)'s name.
        
        Returns
        -------
        value : `None`, `str`
            The value, the user has been typed.
        """
        if not option_names:
            return
            
        option_name, *option_names = option_names
        
        for option in self.iter_options():
            if option.name == option_name:
                return option.get_value_of(*option_names)
    
    
    @property
    def value(self):
        """
        Returns the focused option's value of the application command autocomplete interaction.
        
        Returns
        -------
        value : `None`, `str`
        """
        focused_option = self.focused_option
        if (focused_option is not None):
            return focused_option.value
    
    # Message component
    
    def iter_values(self):
        """
        Iterates over the values selected by the user.
        
        This method is an iterable generator.
        
        Yields
        ------
        value : `str`
        """
        values = self.values
        if (values is not None):
            yield from values
    
    
    def iter_entities(self):
        """
        Iterates over the entities that were selected by the user of a select component interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        entity : ``Channel``, ``ClientUserbase``, ``Role``
        """
        resolved = self.resolved
        if resolved is None:
            return
        
        values = self.values
        if values is None:
            return
        
        try:
            resolver = ENTITY_RESOLVERS[self.component_type]
        except KeyError:
            return
        
        for value in values:
            try:
                entity_id = int(value)
            except ValueError:
                continue
            
            entity = resolver(resolved, entity_id)
            if (entity is not None):
                yield entity
    
    
    @property
    def entities(self):
        """
        Returns the entities that were selected by the user of a select component interaction.
        
        Returns
        -------
        entities : `list` of (``Channel``, ``ClientUserbase``, ``Role``)
        """
        return [*self.iter_entities()]
    
    # Form submit
    
    def iter_components(self):
        """
        Iterates over the sub-components of a form-submit interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``InteractionComponent``
        """
        components = self.components
        if (components is not None):
            yield from components
    
    
    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id : `str`
            The `custom_id` of a represented component.
        value : `str`
            The `value` passed by the user.
        """
        for component in self.iter_components():
            yield from component.iter_custom_ids_and_values()
    
    
    def get_custom_id_value_relation(self):
        """
        Returns a dictionary with `custom_id` to `value` relation.
        
        Returns
        -------
        custom_id_value_relation : `dict` of (`str`, `str`) items
        """
        custom_id_value_relation = {}
        
        for custom_id, value in self.iter_custom_ids_and_values():
            if (value is not None):
                custom_id_value_relation[custom_id] = value
        
        return custom_id_value_relation
    
    
    def get_value_for(self, custom_id_to_match):
        """
        Returns the value for the given `custom_id`.
        
        Parameters
        ----------
        custom_id_to_match : `str`
            A respective components `custom_id` to match.
        
        Returns
        -------
        value : `None`, `str`
            The value if any.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            if (custom_id == custom_id_to_match):
                return value
    
    
    def get_match_and_value(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Returns
        -------
        match : `None`, `Any`
            The returned value by the ``matcher``
        value : `None`, `str`
            The matched `custom_id`'s value.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                return match, value
        
        return None, None
    
    
    def iter_matches_and_values(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Yields
        -------
        match : `None`, `Any`
            The returned value by the ``matcher``
        value : `None`, `str`
            The matched `custom_id`'s value.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                yield match, value
    
    # resolved
    
    @copy_docs(Resolved.resolve_attachment)
    def resolve_attachment(self, attachment_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_attachment(attachment_id)
    
    
    @copy_docs(Resolved.resolve_channel)
    def resolve_channel(self, channel_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_channel(channel_id)
    
    
    @copy_docs(Resolved.resolve_role)
    def resolve_role(self, role_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_role(role_id)
    
    
    @copy_docs(Resolved.resolve_message)
    def resolve_message(self, message_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_message(message_id)
    
    
    @copy_docs(Resolved.resolve_user)
    def resolve_user(self, user_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_user(user_id)
    
    
    @copy_docs(Resolved.resolve_mentionable)
    def resolve_mentionable(self, mentionable_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_mentionable(mentionable_id)
    
    
    @copy_docs(Resolved.resolve_entity)
    def resolve_entity(self, entity_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_entity(entity_id)



__all__ = ('ActivityChange', 'ActivityUpdate',)


class ActivityChange:
    """
    Represents a user's changed activities.
    
    Attributes
    ----------
    added : `None`, `list` of ``Activity``
        The added activities to the respective user. Defaults to `None`.
    updated : `None`, `list` of ``ActivityUpdate``
        The updated activities of the respective user. Defaults to `None`.
    removed: `None`, `list` of ``Activity``
        The removed activities from the respective user. Defaults to `None`.
    """
    __slots__ = ('added', 'updated', 'removed',)
    
    def __init__(self, added, updated, removed):
        """
        Creates a new activity change with the given parameters.
        
        Parameters
        ----------
        added : `None`, `list` of ``Activity``
            The added activities to the user.
        updated : `None`, `list` of ``ActivityUpdate``
            The updated activities of the user.
        removed: `None`, `list` of ``Activity``
            The removed activities from the user.
        """
        self.added = added
        self.updated = updated
        self.removed = removed
    
    def __repr__(self):
        """Returns the representation of the activity change."""
        repr_parts = ['<',
            self.__class__.__name__,
        ]
        
        added = self.added
        if added is None:
            field_added = False
        else:
            repr_parts.append(' added=')
            repr_parts.append(repr(added))
            field_added = True
        
        updated = self.updated
        if (updated is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' updated=')
            repr_parts.append(repr(updated))
        
        removed = self.removed
        if (removed is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' removed=')
            repr_parts.append(repr(removed))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 3
    
    
    def __iter__(self):
        """
        Unpacks the activity change.
        
        This method is a generator.
        """
        yield self.added
        yield self.updated
        yield self.removed
    
    
    def iter_added(self):
        """
        Iterates over the added activities.
        
        This method is an iterable generator.
        
        Yields
        ------
        added : ``Activity``
        """
        added = self.added
        if (added is not None):
            yield from added
    
    
    def iter_updated(self):
        """
        Iterates over the activity updates.
        
        This method is an iterable generator.
        
        Yields
        ------
        updated : ``ActivityUpdate``
        """
        updated = self.updated
        if (updated is not None):
            yield from updated
    
    
    def iter_removed(self):
        """
        Iterates over the removed activities.
        
        This method is an iterable generator.
        
        Yields
        ------
        removed : ``Activity``
        """
        removed = self.removed
        if (removed is not None):
            yield from removed



class ActivityUpdate:
    """
    Represents an updated activity with storing the activity and it's old updated attributes in a `dict`.
    
    Attributes
    ----------
    activity : ``Activity``
        The updated activity.
    old_attributes : `dict` of (`str`, `Any`) items
        The changed attributes of the activity in `attribute-name` - `old-value` relation. Can contain any of the
        following items:
        
        +-------------------+-----------------------------------+
        | Keys              | Values                            |
        +===================+===================================+
        | assets            | `None`, ``ActivityAssets``        |
        +-------------------+-----------------------------------+
        | created_at        | `datetime`                        |
        +-------------------+-----------------------------------+
        | details           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | emoji             | `None`, ``Emoji``                 |
        +-------------------+-----------------------------------+
        | flags             | ``ActivityFlag``                  |
        +-------------------+-----------------------------------+
        | name              | `str`                             |
        +-------------------+-----------------------------------+
        | metadata          | ``ActivityMetadataBase``          |
        +-------------------+-----------------------------------+
        | party             | `None`, ``ActivityParty``         |
        +-------------------+-----------------------------------+
        | secrets           | `None`, ``ActivitySecrets``       |
        +-------------------+-----------------------------------+
        | session_id        | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | state             | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | sync_id           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | timestamps        | `None`, `ActivityTimestamps``     |
        +-------------------+-----------------------------------+
        | url               | `None`, `str`                     |
        +-------------------+-----------------------------------+
    """
    __slots__ = ('activity', 'old_attributes',)
    
    def __init__(self, activity, old_attributes):
        """
        Creates a new activity change instance with the given parameters.
        
        activity : ``Activity``
            The updated activity.
        old_attributes : `dict` of (`str`, `Any`) items
            The changed attributes of the activity.
        """
        self.activity = activity
        self.old_attributes = old_attributes
    
    def __repr__(self):
        """Returns the representation of the activity update."""
        return f'<{self.__class__.__name__} activity={self.activity!r} changes count={len(self.old_attributes)}>'
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 2
    
    def __iter__(self):
        """
        Unpacks the activity update.
        
        This method is a generator.
        """
        yield self.activity
        yield self.old_attributes



__all__ = ()

from .shared import convert_nothing


INTEGRATION_CONVERTERS = {
    'enable_emoticons': convert_nothing,
    'expire_behavior': convert_nothing,
    'expire_grace_period': convert_nothing,
}



__all__ = ()

from ....channel import VoiceRegion
from ....localization import Locale

from ...flags import SystemChannelFlag
from ...preinstanced import ContentFilterLevel, HubType, MFA, MessageNotificationLevel, NsfwLevel, VerificationLevel


from ..audit_log_change import AuditLogChange

from .shared import (
    _convert_preinstanced, convert_deprecated, convert_icon, convert_nothing, convert_snowflake
)


def convert_content_filter(name, data):
    return _convert_preinstanced('content_filter', data, ContentFilterLevel)


def convert_hub_type(name, data):
    return _convert_preinstanced('hub_type', data, HubType)


def convert_bool__boost_progress_bar_enabled(name, data):
    return convert_nothing('boost_progress_bar_enabled', data)


def convert_int__days(name, data):
    return convert_nothing('days', data)


def convert_str__vanity_code(name, data):
    return convert_nothing('vanity_code', data)


def convert_message_notification(name, data):
    return _convert_preinstanced('message_notification', data, MessageNotificationLevel)


def convert_mfa(name, data):
    return _convert_preinstanced('mfa', data, MFA)


def convert_nsfw_level(name, data):
    return _convert_preinstanced('mfa', data, NsfwLevel)


def convert_voice_region(name, data):
    return _convert_preinstanced('region', data, VoiceRegion)


def convert_system_channel_flags(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = SystemChannelFlag(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = SystemChannelFlag(after)
    
    return AuditLogChange('system_channel_flags', before, after)


def convert_verification_level(name, data):
    return _convert_preinstanced('verification_level', data, VerificationLevel)


def convert_preferred_locale(name, data):
    return _convert_preinstanced('preferred_locale', data, Locale)


GUILD_CONVERTERS = {
    'afk_channel_id': convert_snowflake,
    'afk_timeout': convert_nothing,
    'application_id': convert_snowflake,
    'banner_hash': convert_icon,
    'default_message_notifications': convert_message_notification,
    'description': convert_nothing,
    'discovery_splash_hash': convert_icon,
    'explicit_content_filter': convert_content_filter,
    'hub_type': convert_hub_type,
    'icon_hash': convert_icon,
    'mfa_level': convert_mfa,
    'name': convert_nothing,
    'nsfw': convert_deprecated,
    'nsfw_level': convert_nsfw_level,
    'owner_id': convert_snowflake,
    'premium_progress_bar_enabled': convert_bool__boost_progress_bar_enabled,
    'preferred_locale': convert_preferred_locale,
    'prune_delete_days': convert_int__days,
    'public_updates_channel_id': convert_snowflake,
    'rules_channel_id': convert_snowflake,
    'region': convert_voice_region,
    'splash_hash': convert_icon,
    'system_channel_id': convert_snowflake,
    'system_channel_flags': convert_system_channel_flags,
    'vanity_url_code': convert_str__vanity_code,
    'verification_level': convert_verification_level,
    'widget_channel_id': convert_snowflake,
    'widget_enabled': convert_nothing,
}



__all__ = ('CommandFunction',)

from scarletio import RichAttributeErrorBaseType

from .parameter import get_command_parameters_for
from .parameter_renderer import (
    ParameterRenderer, get_lengths, get_render_generic_line_length, get_render_modifier_line_length
)
from .parameter_result import ParameterResult
from .render_constants import (
    BOX_LEFT, BOX_RIGHT, BOX_TITLE_MODIFIER, BOX_TITLE_PARAMETER, BOX_TITLE_SUB_COMMANDS, NOTE_SIGN_DESCRIPTION
)
from .rendering_helpers import (
    render_box_end_into, render_box_line_adjustment_into, render_box_start_into, render_sub_command_box_into,
    render_usage_line_into
)
from .result import (
    COMMAND_RESULT_CODE_CALL, COMMAND_RESULT_CODE_PARAMETER_REQUIRED, COMMAND_RESULT_CODE_PARAMETER_UNEXPECTED,
    COMMAND_RESULT_CODE_PARAMETER_UNSATISFIED, CommandResult
)


class CommandFunction(RichAttributeErrorBaseType):
    """
    Command line function to call.
    
    Attributes
    ----------
    _function : `callable`
        The function to call.
    _parent_reference : `None`, ``WeakReferer``
        Weakreference to the command function's parent.
    _parameters : `list` of ``CommandParameter``
        Parameter parsers to call the command with.
    description : `None`, `str`
        The command's description.
    name : `str`
        The command's name.
    """
    __slots__ = (
        '_function', '_parent_reference', '_parameters', 'description', 'name',
    )
    
    def __new__(cls, parent, function, name, description):
        """
        Creates a new ``CommandFunction``.
        
        Parameters
        ----------
        parent : ``CommandCategory``
            The command's parent.
        function : `callable`
            The function to call when the command is used.
        name : `str`
            The command's name.
        description : `None`, `str`
            The command's description.
        """
        parameters = get_command_parameters_for(function)
        
        self = object.__new__(cls)
        self._function = function
        self._parameters = parameters
        self._parent_reference = parent._self_reference
        self.description = description
        self.name = name
        return self
    
    
    def __repr__(self):
        """Returns the command function's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name=')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_parent(self):
        """
        Returns the parent category of the command.
        
        Returns
        -------
        parent : `None`, ``CommandCategory``
        """
        parent_reference = self._parent_reference
        if (parent_reference is not None):
            return parent_reference()
    
    
    def iter_direct_sub_command_names(self):
        """
        gets the direct sub-command names of the command function.
        
        This method is an iterable generator.
        
        Yields
        ------
        sub_command_name : `str`
        """
        parent = self.get_parent()
        if (parent is not None):
            yield from parent.iter_sub_category_names()
    
    
    def _trace_back_name(self):
        """
        Traces back to the source name of the command.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
        
        Returns
        -------
        parent_name : `None`
        """
        parent_name = None
        
        parent = self.get_parent()
        if (parent is not None):
            parent_name = yield from parent._trace_back_name()
        
        name = self.name
        if (parent_name is None) or (parent_name != name):
            yield name
        
        return None
    
    
    def invoke(self, parameters, index):
        """
        Calls the command line command.
        
        Parameters
        ----------
        parameters : `list` of `str`
            Command line parameters.
        index : `int`
            The index of the first parameter trying to process.
        
        Returns
        -------
        command_result : ``CommandResult``
        """
        positional_parameters = []
        keyword_parameters = {}
        
        command_result = parse_parameters_into(
            parameters, index, self._parameters, positional_parameters, keyword_parameters
        )
        if (command_result is not None):
            return command_result
        
        
        return CommandResult(
            COMMAND_RESULT_CODE_CALL,
            self,
            positional_parameters,
            keyword_parameters,
        )
    
    
    def get_usage(self):
        """
        Returns the usage of the command function.
        
        Returns
        -------
        usage : `str`
        """
        return ''.join(self.render_usage_into([]))
    
    
    def render_usage_into(self, into):
        """
        Renders the command function's usage into the given list.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the usage into.
        
        Returns
        -------
        into : `list` of `str`
        """
        
        # Display description
        
        description = self.description
        if (description is not None):
            into.append(description)
            into.append('\n\n')
        
        into = render_usage_line_into(into, self._trace_back_name())
        
        parameter_renderers = [ParameterRenderer(command_parameter) for command_parameter in self._parameters]
        
        modifier_field_added = False
        should_add_rest_field = False
        
        for parameter_renderer in parameter_renderers:
            if parameter_renderer.is_positional():
                into.append(' ')
                into.append(parameter_renderer.get_name().upper())
                continue
            
            if parameter_renderer.is_args():
                into.append(' [')
                into.append(parameter_renderer.get_name().upper())
                into.append('...]')
                continue
            
            if parameter_renderer.is_modifier():
                if modifier_field_added:
                    continue
                
                into.append(' [OPTIONS]')
                modifier_field_added = True
                continue
            
            should_add_rest_field = True
        
        if should_add_rest_field:
            into.append(' ...')
        
        parameters_generic = []
        parameters_modifier = []
        
        for parameter_renderer in parameter_renderers:
            if parameter_renderer.is_modifier():
                parameters_modifier.append(parameter_renderer)
            else:
                parameters_generic.append(parameter_renderer)
        
        sub_command_names = [*self.iter_direct_sub_command_names()]
        
        generic_lengths = get_lengths(parameters_generic, ParameterRenderer.get_generic_lengths)
        generic_line_length = get_render_generic_line_length(generic_lengths)
        modifier_lengths = get_lengths(parameters_modifier, ParameterRenderer.get_modifier_lengths)
        modifier_line_length = get_render_modifier_line_length(modifier_lengths)
        sub_command_line_length = max((len(sub_command_name) for sub_command_name in sub_command_names), default = 0)
        
        
        line_length = 0
        if generic_line_length:
            line_length = max(line_length, generic_line_length, len(BOX_TITLE_PARAMETER))
        if modifier_line_length:
            line_length = max(line_length, modifier_line_length, len(BOX_TITLE_MODIFIER))
        if sub_command_line_length:
            line_length = max(line_length, sub_command_line_length, len(BOX_TITLE_SUB_COMMANDS))
        
        field_added = False
        
        if generic_line_length:
            if not field_added:
                into.append('\n\n')
                field_added = True
            
            # Render line 1
            
            into = render_box_start_into(into, line_length, BOX_TITLE_PARAMETER)
            
            # Render line n
            for parameter in parameters_generic:
                into.append(BOX_LEFT)
                into = parameter.render_generic_into_with_lengths(into, generic_lengths)
                into = render_box_line_adjustment_into(into, generic_line_length, line_length)
                into.append(BOX_RIGHT)
                into.append('\n')
            
            # Render line -1
            into = render_box_end_into(into, line_length)
        
        
        if modifier_line_length:
            if not field_added:
                into.append('\n\n')
                field_added = True
            
            # Render line 1
            into = render_box_start_into(into, line_length, BOX_TITLE_MODIFIER)
            
            # Render line n
            for parameter in parameters_modifier:
                into.append(BOX_LEFT)
                into = parameter.render_modifier_into_with_lengths(into, modifier_lengths)
                into = render_box_line_adjustment_into(into, modifier_line_length, line_length)
                into.append(BOX_RIGHT)
                into.append('\n')
            
            # Render line -1
            into = render_box_end_into(into, line_length)
        
        
        if sub_command_line_length:
            if not field_added:
                into.append('\n\n')
                field_added = True
            
            into = render_sub_command_box_into(into, sub_command_names, line_length)
        
        
        # Collect signs
        note_signs = None
        
        for parameter_renderer in parameter_renderers:
            note_sign = parameter_renderer.get_note_sign()
            if (note_sign is not None):
                if note_signs is None:
                    note_signs = set()
                
                note_signs.add(note_sign)
        
        # Display note signs
        if (note_signs is not None):
            if not field_added:
                into.append('\n')
                field_added = True
            
            note_sign_adjustment = max(len(note_sign) for note_sign in note_signs) + 1
            for note_sign in sorted(note_signs):
                into.append('\n')
                into.append(note_sign)
                into.append(' ' * (note_sign_adjustment - len(note_sign)))
                into.append(NOTE_SIGN_DESCRIPTION[note_sign])
        
        return into
    
    
    def get_full_name(self):
        """
        Returns the command function's full name.
        
        Returns
        -------
        full_name : `str`
        """
        return ' '.join(self._trace_back_name())


def parse_parameters_into(parameter_values, start_index, command_parameters, positional_parameters, keyword_parameters):
    """
    Parameters
    ----------
    parameter_values : `list` of `str`
        Command line parameters.
    start_index : `int`
        The index of the first parameter trying to process.
    command_parameters : `list` of ``CommandParameter``
        Parameters of the respective command.
    positional_parameters : `list` of `Any`
        Positional parameters to call the respective function with.
    keyword_parameters : `dict` of (`str`, `Any`) items
        Keyword parameters to call the respective function with.
    
    Returns
    -------
    command_result : `None`, ``CommandResult``
        Returns a command result if failed.
    """
    parameter_values = parameter_values[start_index:]
    
    for parameter_value in parameter_values:
        if parameter_value.startswith('---'):
            return CommandResult(
                COMMAND_RESULT_CODE_PARAMETER_UNEXPECTED,
                parameter_value,
            )
    
    parameter_values.reverse()
    
    
    parameter_results = [ParameterResult(command_parameter) for command_parameter in command_parameters]
    
    while parameter_values:
        parameter_value = parameter_values.pop()
        
        if parameter_value.startswith('--'):
            parameter_name, parsed_value = parse_modifier_parameter_name(parameter_value)
            parameter_result = find_satisfiable_modifier_parameter_result_for(parameter_results, parameter_name)
            if (parameter_result is None):
                return CommandResult(
                    COMMAND_RESULT_CODE_PARAMETER_UNEXPECTED,
                    parameter_value,
                )
            
            command_result = parameter_result.feed(parsed_value)
            if (command_result is not None):
                return command_result
            
            continue
        
        
        if parameter_value.startswith('-'):
            parameter_name = re_normalise_parameter_name(parameter_value)
            parameter_result = find_satisfiable_keyword_parameter_result_for(
                parameter_results,
                parameter_name,
            )
            
            if (parameter_result is None):
                return CommandResult(
                    COMMAND_RESULT_CODE_PARAMETER_UNEXPECTED,
                    parameter_value,
                )
            
            if len(parameter_values) == 0:
                return CommandResult(
                    COMMAND_RESULT_CODE_PARAMETER_UNSATISFIED,
                    parameter_result.command_parameter,
                )
        
            command_result = parameter_result.feed_as(parameter_values.pop(), parameter_name)
            if (command_result is not None):
                return command_result
            
            continue
        
        
        parameter_result = find_next_satisfiable_positional_parameter_result(parameter_results)
        if (parameter_result is None):
            return CommandResult(
                COMMAND_RESULT_CODE_PARAMETER_UNEXPECTED,
                parameter_value,
            )
        
        command_result = parameter_result.feed(parameter_value)
        if (command_result is not None):
            return command_result
        
        continue
    
    
    for parameter_result in parameter_results:
        if not parameter_result.put_into(positional_parameters, keyword_parameters):
            return CommandResult(
                COMMAND_RESULT_CODE_PARAMETER_REQUIRED,
                parameter_result.command_parameter,
            )
    
    return None


def re_normalise_parameter_name(parameter_name):
    """
    Re-normalises parameter name.
    
    Parameters
    ----------
    parameter_name : `str`
        The parameter name to normalise.
    
    Returns
    -------
    parameter_name : `str`
    """
    return '-'.join(parameter_name.replace('_', ' ').replace('-', ' ').split())


def parse_modifier_parameter_name(parameter_name):
    """
    Parses modifier type parameter.
    
    Parameters
    ----------
    parameter_name : `str`
        The parameter name to parse.
    
    Returns
    -------
    output_name : `str`
        The parameter's normalises and stripped name.
    output_value : `bool`
        The set value to the parameter.
    """
    parameter_name = re_normalise_parameter_name(parameter_name)
    if parameter_name.startswith('no-'):
        output_value = False
        output_name = parameter_name[len('no-'):]
    else:
        output_value = True
        output_name = parameter_name
    
    return output_name, output_value


def find_satisfiable_modifier_parameter_result_for(parameter_results, parameter_name):
    """
    Finds a satisfiable modifier parameter for the given nane.
    
    Parameters
    ----------
    parameter_results : `list` of ``ParameterResult``
        Parameter results to select from.
    parameter_name : `str`
        Parameter name to match.
    
    Returns
    -------
    parameter_result : ``ParameterResult``
        The matched parameter parse result.
    """
    for parameter_result in parameter_results:
        if (
            parameter_result.is_satisfiable() and
            parameter_result.matches_name(parameter_name) and
            parameter_result.is_modifier()
        ):
            return parameter_result


def find_satisfiable_keyword_parameter_result_for(parameter_results, parameter_name):
    """
    Finds a satisfiable keyword parameter for the given nane.
    
    Parameters
    ----------
    parameter_results : `list` of ``ParameterResult``
        Parameter results to select from.
    parameter_name : `str`
        Parameter name to match.
    
    Returns
    -------
    parameter_result : ``ParameterResult``
        The matched parameter parse result.
    """
    for parameter_result in parameter_results:
        if (
            parameter_result.is_satisfiable() and
            parameter_result.matches_name(parameter_name)
        ):
            return parameter_result


def find_next_satisfiable_positional_parameter_result(parameter_results):
    """
    Finds a satisfiable positional parameter for the given nane.
    
    Parameters
    ----------
    parameter_results : `list` of ``ParameterResult``
        Parameter results to select from.
    
    Returns
    -------
    parameter_result : ``ParameterResult``
        The matched parameter parse result.
    """
    for parameter_result in parameter_results:
        if (
            parameter_result.is_satisfiable() and
            parameter_result.is_positional()
        ):
            return parameter_result



__all__ = ('IntegrationMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder

from ..integration_account import IntegrationAccount

from .constants import EXPIRE_GRACE_PERIOD_DEFAULT, SUBSCRIBER_COUNT_DEFAULT
from .preinstanced import IntegrationExpireBehavior


class IntegrationMetadataBase(RichAttributeErrorBaseType):
    """
    Base type specific metadata for integrations.
    """
    __all__ = ()
    
    account = PlaceHolder(
        IntegrationAccount._create_empty(),
        """
        The integration's respective account. If the integration type is `'discord'`, then set as a discord user
        itself.
        
        Returns
        -------
        account : ``IntegrationAccount``, ``ClientUserBase``
        """
    )
    
    
    application = PlaceHolder(
        None,
        """
        The application of the integration if applicable.
        
        Returns
        -------
        application : ``IntegrationApplication``
        """
    )
    
    
    expire_behavior = PlaceHolder(
        IntegrationExpireBehavior.remove_role,
        """
        The behavior of expiring subscription.
        
        Returns
        -------
        expire_behavior : ``IntegrationExpireBehavior``
        """
    )
    
    
    expire_grace_period = PlaceHolder(
        EXPIRE_GRACE_PERIOD_DEFAULT,
        """
        The grace period in days for expiring subscribers.
        
        Returns
        -------
        expire_grace_period : `int`
        """
    )
    
    
    revoked = PlaceHolder(
        False,
        """
        Whether the integration is removed.
        
        Returns
        -------
        revoked : `bool`
        """
    )
    
    
    role_id = PlaceHolder(
        0,
        """
        The role's identifier what the integration uses for subscribers.
        
        Returns
        -------
        role_id : `int`
        """
    )
    
    
    scopes = PlaceHolder(
        None,
        """
        The scopes the application was authorized with.
        
        Returns
        -------
        scopes : `None`, `tuple` of ``Oauth2Scope``
        """
    )
    
    
    subscriber_count = PlaceHolder(
        SUBSCRIBER_COUNT_DEFAULT,
        """
        How many subscribers the integration has. 
        
        Returns
        -------
        subscriber_count : `int`
        """
    )
    
    
    synced_at = PlaceHolder(
        None,
        """
        When the integration was last synced.
        
        synced_at : `None`, `datetime`
        """
    )
    
    
    syncing = PlaceHolder(
        False,
        """
        Whether the integration syncing.
        
        Returns
        -------
        syncing : `bool`
        """
    )
    
    def __eq__(self, other):
        """Returns whether the two integration accounts are the same."""
        if type(self) is not type(other):
            return False
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two integration metadatas are equal.
        
        > Both object must have the same type.
        
        Parameters
        ----------
        other : `instance<type<cls>>`
            The other instance.
        """
        return True
    
    
    def __hash__(self):
        """Returns the integration account's hash value."""
        return 0
    
    
    def __repr__(self):
        return f'<{self.__class__.__name__}>'
    
    
    def __new__(cls, keyword_parameters):
        """
        Creates a new integration metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters to populate the attributes.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an empty integration metadata with it's default attributes set.
        
        Returns
        -------
        self : `instance<self>`
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new integration metadata from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Integration data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the integration metadata into json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}



__all__ = ()

from scarletio.tools.asynchronous_interactive_console import (
    collect_package_local_variables, create_banner, create_exit_message, run_asynchronous_interactive_console
)
from scarletio.tools.asynchronous_interactive_console.console_helpers import AWAIT_NOTE

from .... import __package__ as PACKAGE_NAME

from ... import register


PACKAGE = __import__(PACKAGE_NAME)


@register(
    alters = 'i',
    description =  f'Runs asynchronous python interpreter through scarletio.\n{AWAIT_NOTE}',
)
def interpreter():
    run_asynchronous_interactive_console(
        collect_package_local_variables(PACKAGE),
        banner = create_banner(PACKAGE),
        exit_message = create_exit_message(PACKAGE),
    )



__all__ = ('ChannelMetadataGuildAnnouncements',)

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import (
    PERMISSION_DENY_SEND_MESSAGES_IN_THREADS_ONLY, PERMISSION_MASK_MANAGE_MESSAGES, PERMISSION_MASK_SEND_MESSAGES,
    PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_TEXT_DENY, PERMISSION_VOICE_DENY
)

from .guild_text_base import ChannelMetadataGuildTextBase


class ChannelMetadataGuildAnnouncements(ChannelMetadataGuildTextBase):
    """
    Guild announcements channel metadata.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    default_thread_auto_archive_after : `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Defaults
        to `3600`. Can be one of: `3600`, `86400`, `259200`, `604800`.
    default_thread_slowmode : `int`
        Applied as `thread.slowmode` when one is created.
    nsfw : `bool`
        Whether the channel is marked as non safe for work.
    topic : `None`, `str`
        The channel's topic.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channel` permissions are unaffected.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()
    
    @copy_docs(ChannelMetadataGuildTextBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # text channels don't have voice permissions
        result &= PERMISSION_VOICE_DENY
        
        if result & PERMISSION_MASK_MANAGE_MESSAGES:
            if result & PERMISSION_MASK_SEND_MESSAGES:
                result &= PERMISSION_DENY_SEND_MESSAGES_IN_THREADS_ONLY
            else:
                result &= PERMISSION_TEXT_DENY
        else:
            result &= PERMISSION_TEXT_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildTextBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # text channels don't have voice permissions
        result &= PERMISSION_VOICE_DENY
        
        if result & PERMISSION_MASK_MANAGE_MESSAGES:
            if result & PERMISSION_MASK_SEND_MESSAGES:
                result &= PERMISSION_DENY_SEND_MESSAGES_IN_THREADS_ONLY
            else:
                result &= PERMISSION_TEXT_DENY
        else:
            result &= PERMISSION_TEXT_DENY
        
        return Permission(result)



__all__ = ('PLUGINS',)

from scarletio import WeakValueDictionary


PLUGINS = WeakValueDictionary()

PLUGIN_STATE_UNDEFINED = 0
PLUGIN_STATE_LOADED = 1
PLUGIN_STATE_UNLOADED = 2
PLUGIN_STATE_UNSATISFIED = 3

PLUGIN_STATE_VALUE_TO_NAME = {
    PLUGIN_STATE_UNDEFINED: 'undefined',
    PLUGIN_STATE_LOADED: 'loaded',
    PLUGIN_STATE_UNLOADED: 'unloaded',
    PLUGIN_STATE_UNSATISFIED: 'unsatisfied',
}

LOADING_PLUGINS = set()

# Plugin roots define core part for plugins.
PLUGIN_ROOTS = set()


ABSOLUTE_PATH_PLUGIN_NAME_PREFIX = '<extension>.'



__all__ = ('AutoModerationRuleTriggerMetadataMentionSpam',)

from scarletio import copy_docs

from ..constants import AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX

from .base import AutoModerationRuleTriggerMetadataBase


class AutoModerationRuleTriggerMetadataMentionSpam(AutoModerationRuleTriggerMetadataBase):
    """
    Mention spam trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    mention_limit : `int`
        The amount of mentions in a message after the rule is triggered.
    """
    __slots__ = ('mention_limit',)
    
    def __new__(cls, mention_limit):
        """
        Creates a new mention spam trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        mention_limit : `None`, `int`
            The amount of mentions in a message after the rule is triggered.
            
            Defaults to the allowed maximal amount if given as `None`.
            
        Raises
        ------
        TypeError
            - If `mention_limit` type is incorrect.
        """
        if (mention_limit is None):
            mention_limit = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
        
        elif isinstance(mention_limit, int):
            if mention_limit < 0:
                mention_limit = 0
            
            elif mention_limit > AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX:
                mention_limit = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
            
        else:
            raise TypeError(
                f'`mention_limit` can be `None`, `int`, got {mention_limit.__class__.__name__}; {mention_limit!r}.'
            )
        
        self = object.__new__(cls)
        self.mention_limit = mention_limit
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # mention_limit
        repr_parts.append(' mention_limit=')
        repr_parts.append(repr(self.mention_limit))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadataBase.from_data)
    def from_data(cls, data):
        mention_limit = data.get('mention_total_limit', None)
        if (mention_limit is None):
            mention_limit = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
        
        self = object.__new__(cls)
        self.mention_limit = mention_limit
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.to_data)
    def to_data(self):
        data = {}
        
        data['mention_total_limit'] = self.mention_limit
        
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.mention_limit != other.mention_limit:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # mention_limit
        hash_value ^= self.mention_limit
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.copy)
    def copy(self):
        new = AutoModerationRuleTriggerMetadataBase.copy(self)
        
        # mention_limit
        new.mention_limit = self.mention_limit
        
        return new



__all__ = ('ApplicationCommandPermission',)

from scarletio import RichAttributeErrorBaseType

from ..bases import maybe_snowflake

from .application_command import ApplicationCommand
from .application_command_permission_overwrite import ApplicationCommandPermissionOverwrite
from .constants import APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX


class ApplicationCommandPermission(RichAttributeErrorBaseType):
    """
    Stores am ``ApplicationCommand``'s overwrites.
    
    Attributes
    ----------
    application_command_id : `int`
        The identifier of the respective ``ApplicationCommand``.
    
    application_id : `int`
        The application command's application's identifier.
    
    guild_id : `int`
        The identifier of the respective guild.
    
    permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The application command permissions overwrites relating to the respective application command in the guild.
    """
    __slots__ = ('application_command_id', 'application_id', 'guild_id', 'permission_overwrites')
    
    def __new__(cls, application_command, *, permission_overwrites=None):
        """
        Creates a new ``ApplicationCommandPermission`` from the given parameters.
        
        Parameters
        ----------
        application_command : ``ApplicationCommand``, `int`
            The application command's identifier.
        
        permission_overwrites : `None`, (`list`, `set`, `tuple`) of ``ApplicationCommandPermissionOverwrite` = `None`
                , Optional (Keyword only)
            Overwrites for the application command.
        
        Raises
        ------
        TypeError
            - If `application_command` was not given neither as ``ApplicationCommand`` nor as `int`.
        AssertionError
            - If `permission_overwrites` was not give neither as `None`, `list`, `set`, `tuple`.
            - If `permission_overwrites` contains a non ``ApplicationCommandPermissionOverwrite`` element.
            - If `permission_overwrites` length is over `10`.
        """
        # application_command
        if isinstance(application_command, ApplicationCommand):
            application_command_id = application_command.id
        else:
            application_command_id = maybe_snowflake(application_command)
            if application_command_id is None:
                raise TypeError(
                    f'`application_command` can be `{ApplicationCommand.__name__}`, `int` , got '
                    f'{application_command.__class__.__name__}; {application_command!r}.'
                )
        
        # application_id
        # Internal attribute
        
        # guild_id
        # Internal attribute
        
        # permission_overwrites
        if permission_overwrites is None:
            permission_overwrites_processed = None
        
        else:
            if __debug__:
                if not isinstance(permission_overwrites, (list, set, tuple)):
                    raise AssertionError(
                        f'`permission_overwrites` can be `None`, `list`, `set`, `tuple`, got '
                        f'{permission_overwrites.__class__.__name__}; {permission_overwrites!r}.'
                    )
            
            permission_overwrites_processed = []
            
            for permission_overwrite in permission_overwrites:
                if __debug__:
                    if not isinstance(permission_overwrite, ApplicationCommandPermissionOverwrite):
                        raise AssertionError(
                            f'`permission_overwrites` contains a non '
                            f'`{ApplicationCommandPermissionOverwrite.__name__}` element, got '
                            f'{permission_overwrite.__class__.__name__}; {permission_overwrite!r}; '
                            f'permission_overwrites={permission_overwrites!r}.'
                        )
                
                permission_overwrites_processed.append(permission_overwrite)
                
            
            if permission_overwrites_processed:
                if __debug__:
                    if len(permission_overwrites_processed) >= APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX:
                        raise AssertionError(
                            f'`permission_overwrites` can have up to '
                            f'`{APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX}` elements, which is already reached, '
                            f'got {permission_overwrites_processed!r}.'
                        )
            else:
                permission_overwrites_processed = None
        
        self = object.__new__(cls)
        self.application_command_id = application_command_id
        self.application_id = 0
        self.guild_id = 0
        self.permission_overwrites = permission_overwrites_processed
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandPermission``.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Application command data.
        
        Returns
        -------
        self : ``ApplicationCommandPermission``
        """
        # application_command_id
        application_command_id = int(data['id'])
        
        # application_id
        application_id = int(data['application_id'])
        
        # guild_id
        guild_id = int(data['guild_id'])
        
        # permission_overwrites
        permission_overwrite_datas = data['permissions']
        if permission_overwrite_datas:
            permission_overwrites = [
                ApplicationCommandPermissionOverwrite.from_data(permission_overwrite_data) for
                permission_overwrite_data in permission_overwrite_datas
            ]
            
        else:
            permission_overwrites = None
        
        
        self = object.__new__(cls)
        
        self.application_command_id = application_command_id
        self.application_id = application_id
        self.guild_id = guild_id
        self.permission_overwrites = permission_overwrites
        
        return self
    
    
    def to_data(self):
        """
        Converts the application command permission to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # application_command_id
        data['id'] = self.application_command_id
        
        # application_id
        data['application_id'] = self.application_id
        
        # guild_id
        data['guild_id'] = self.guild_id
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if permission_overwrites is None:
            permission_overwrite_datas = []
        else:
            permission_overwrite_datas = [
                permission_overwrite.to_data() for permission_overwrite in permission_overwrites
            ]
        
        data['permissions'] = permission_overwrite_datas
        
        return data
    
    
    def __repr__(self):
        """Returns the application command permission's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # application_command_id
        repr_parts.append(' application_command_id=')
        repr_parts.append(repr(self.application_command_id))
        
        # application_id
        # Unique by application_command_id
        
        # guild_id
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if permission_overwrites is None:
            permission_overwrite_count = '0'
        else:
            permission_overwrite_count = repr(len(permission_overwrites))
        repr_parts.append(', permission overwrite count=')
        repr_parts.append(permission_overwrite_count)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application command permission's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # application_command_id
        if self.application_command_id != other.application_command_id:
            return False
        
        # application_id
        # Unique by application_command_id
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # permission_overwrites
        if self.permission_overwrites != other.permission_overwrites:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the application command overwrite's hash value."""
        hash_value = 0
        
        # application_command_id
        hash_value ^= self.application_command_id
        
        # application_id
        # Unique by application_command_id
        
        # guild_id
        hash_value ^= self.guild_id
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            hash_value ^= len(permission_overwrites) << 4
            
            for permission_overwrite in permission_overwrites:
                hash_value ^= hash(permission_overwrite)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the application command permission.
        
        Returns
        -------
        new : ``ApplicationCommandPermission``
        """
        new = object.__new__(type(self))
        
        # application_command_id
        new.application_command_id = self.application_command_id
        
        # application_id
        new.application_id = self.application_id
        
        # guild_id
        new.guild_id = self.guild_id
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            permission_overwrites = [permission_overwrite.copy() for permission_overwrite in permission_overwrites]
        new.permission_overwrites = permission_overwrites
        
        return new
    
    
    def add_permission_overwrite(self, permission_overwrite):
        """
        Adds an application command permission overwrite to the overwrites of the application command permission.
        
        Parameters
        ----------
        permission_overwrite : ``ApplicationCommandPermissionOverwrite``
            The overwrite to add.
        
        Raises
        ------
        AssertionError
            - If `overwrite` is not ``ApplicationCommandPermissionOverwrite``.
            - If the application command permission has `10` overwrites already.
        """
        if __debug__:
            if not isinstance(permission_overwrite, ApplicationCommandPermissionOverwrite):
                raise AssertionError(
                    f'`permission_overwrite` can be {ApplicationCommandPermissionOverwrite.__name__} , got '
                    f'{permission_overwrite.__class__.__name__}; {permission_overwrite!r}.'
                )
        
        permission_overwrites = self.permission_overwrites
        if permission_overwrites is None:
            self.permission_overwrites = permission_overwrites = []
        else:
            if __debug__:
                if len(permission_overwrites) >= APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX:
                    raise AssertionError(
                        f'`permission_overwrites` can have up to '
                        f'`{APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX}` elements, which is already reached, '
                        f'got {permission_overwrite!r}.'
                    )
        
        permission_overwrites.append(permission_overwrite)
